@@ -14,8 +14,7 @@
 # Public License for more details.
 #
 # You should have received a copy of the GNU General Public License along with
-# this program; if not, write to the Free Software Foundation, Inc.,
-# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA
+# this program. If not, see <https://www.gnu.org/licenses/>
 
 
 #
@@ -24,8 +23,11 @@
 # in a configuration file, that's why).
 #
 
-import ConfigParser
-import UserDict
+try:
+    import UserDict
+except ImportError:
+    from collections import UserDict
+
 import subprocess
 import collections
 import re
@@ -33,6 +35,11 @@ import distro_info
 from functools import reduce
 
 
+import six
+from six.moves import configparser
+from six.moves import reduce
+
+
 class MissingSection(Exception):
 
     def __init__(self, filename, section):
@@ -47,18 +54,18 @@ class MissingMandatorySetting(Exception):
             (key, filename),
 
 
-class Config(UserDict.UserDict):
+class Config(UserDict):
 
     def __init__(self, section, defaults, mandatory=[], defaults_section=None):
-        UserDict.UserDict.__init__(self)
+        UserDict.__init__(self)
         self._section = section
         self._defaults_section = defaults_section
-        for key, value in defaults.iteritems():
+        for key, value in six.iteritems(defaults):
             self[key] = value
         self._mandatory = mandatory
 
     def read(self, filename):
-        cp = ConfigParser.ConfigParser()
+        cp = configparser.ConfigParser()
         cp.read(filename)
         if not cp.has_section(self._section):
             raise MissingSection(filename, self._section)
@@ -73,7 +80,7 @@ class Config(UserDict.UserDict):
     def get_mirror(self, distro=None):
         if self["mirror"] is not None:
             return self["mirror"]
-        return "http://httpredir.debian.org/debian"
+        return "http://deb.debian.org/debian"
 
     def get_distros(self):
         if self["upgrade-test-distros"] is not None:
@@ -114,7 +121,7 @@ class Config(UserDict.UserDict):
         ])
 
         # add mappings for e.g. "oldstable" -> "oldstable"
-        distmap.update(dict([(val, val) for key, val in distmap.iteritems()]))
+        distmap.update(dict([(val, val) for key, val in six.iteritems(distmap)]))
 
         # map e.g. "Debian6" -> "oldstable" where debdist.old(result="fullname")
         # currently returns 'Debian 6.0 "Squeeze"'
@@ -149,10 +156,10 @@ class Config(UserDict.UserDict):
         return self["arch"]
 
 
-class DistroConfig(UserDict.UserDict):
+class DistroConfig(UserDict):
 
     def __init__(self, filename, mirror):
-        UserDict.UserDict.__init__(self)
+        UserDict.__init__(self)
         self._mirror = mirror
         self._defaults = {
             "uri": None,
@@ -162,7 +169,7 @@ class DistroConfig(UserDict.UserDict):
                 "depends": None,
                 "candidates": None,
         }
-        cp = ConfigParser.SafeConfigParser()
+        cp = configparser.SafeConfigParser()
         cp.read(filename)
         for section in cp.sections():
             self[section] = dict(self._defaults)
@@ -247,12 +254,14 @@ class DistroConfig(UserDict.UserDict):
                         c))
         return lines
 
-    def get_basetgz(self, distro, arch):
+    def get_basetgz(self, distro, arch, merged_usr=False):
         # look for the first base distribution
         for d in self._expand_depends(distro):
             if self.get(d, "depends"):
                 next  # skip partial distro
-            return "%s_%s.tar.gz" % (self.get_distribution(d), arch)
+            return "%s%s_%s.tar.gz" % (self.get_distribution(d),
+                                       "-merged-usr" if merged_usr else "",
+                                       arch)
         return None
 
 
