@@ -1,7 +1,7 @@
 # -*- coding: utf-8 -*-
 
 # Copyright 2005 Lars Wirzenius (liw@iki.fi)
-# Copyright © 2011-2014 Andreas Beckmann (anbe@debian.org)
+# Copyright © 2011-2018 Andreas Beckmann (anbe@debian.org)
 #
 # This program is free software; you can redistribute it and/or modify it
 # under the terms of the GNU General Public License as published by the
@@ -14,8 +14,7 @@
 # Public License for more details.
 #
 # You should have received a copy of the GNU General Public License along with
-# this program; if not, write to the Free Software Foundation, Inc.,
-# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA
+# this program. If not, see <https://www.gnu.org/licenses/>
 
 
 """Packages database for distributed piuparts processing
@@ -33,12 +32,18 @@ import random
 import stat
 import tempfile
 import time
-import UserDict
+try:
+    import UserDict
+except ImportError:
+    from collections import UserDict
 import apt_pkg
 
 import piupartslib
 from piupartslib.dependencyparser import DependencyParser
 
+import six
+
+
 apt_pkg.init_system()
 
 
@@ -55,10 +60,10 @@ def rfc822_like_header_parse(input):
     return headers
 
 
-class Package(UserDict.UserDict):
+class Package(UserDict):
 
     def __init__(self, headers):
-        UserDict.UserDict.__init__(self)
+        UserDict.__init__(self)
         self.headers = headers
         for header in headers:
             name, value = header.split(":", 1)
@@ -70,6 +75,43 @@ class Package(UserDict.UserDict):
         self.waiting_cnt = None
         self.rdep_chain_len = None
 
+    def name(self):
+        return self["Package"]
+
+    def version(self):
+        return self["Version"]
+
+    def source(self):
+        # Binary packages built from the source package with the same name
+        # and version don't have a Source header.
+        if "Source" in self:
+            # If source and binary version differ (e.g. for binNMUs), the
+            # source version is given as the second element in the "Source"
+            # entry. Strip off the optional source version.
+            return self["Source"].split(" ")[0]
+        return self["Package"]
+
+    def source_version(self):
+        # Binary packages built from the source package with the same name
+        # and version don't have a Source header.
+        if "Source" in self:
+            # If source and binary version differ (e.g. for binNMUs), the
+            # source version is given as the second element in the "Source"
+            # entry. Strip off the parentheses around the source version.
+            if " " in self["Source"]:
+                return self["Source"].split(" ")[1][1:-1]
+        return self["Version"]
+
+    def set_test_versions(self, tv):
+        self["TestVersions"] = tv
+
+    def test_versions(self):
+        """ToDo: for distupgrade tests test_versions() should return a list
+           of versions in all distros in the upgrade path"""
+        if "TestVersions" in self:
+            return self["TestVersions"]
+        return self.version()
+
     def _parse_dependencies(self, header_name):
         if header_name in self._parsed_deps:
             depends = self._parsed_deps[header_name]
@@ -127,10 +169,10 @@ class Package(UserDict.UserDict):
         output_file.write("".join(self.headers))
 
 
-class PackagesFile(UserDict.UserDict):
+class PackagesFile(UserDict):
 
     def __init__(self):
-        UserDict.UserDict.__init__(self)
+        UserDict.__init__(self)
         self._urllist = []
 
     def load_packages_urls(self, urls, restrict_packages=None):
@@ -194,22 +236,25 @@ class LogDB:
             if basename.endswith(".log"):
                 cache[os.path.join(dirname, basename)] = True
 
-    def open_file(self, pathname, mode):
-        return file(pathname, mode)
-
     def remove_file(self, pathname):
         os.remove(pathname)
 
     def _log_name(self, package, version):
         return "%s_%s.log" % (package, version)
 
-    def log_exists(self, package, subdirs):
-        log_name = self._log_name(package["Package"], package["Version"])
+    def _kpr_name(self, package, version):
+        return "%s_%s.kpr" % (package, version)
+
+    def log_exists2(self, package, version, subdirs):
+        log_name = self._log_name(package, version)
         for subdir in subdirs:
             if self.exists(os.path.join(subdir, log_name)):
                 return True
         return False
 
+    def log_exists(self, package, subdirs):
+        return self.log_exists2(package.name(), package.test_versions(), subdirs)
+
     def create(self, subdir, package, version, contents):
         (fd, temp_name) = tempfile.mkstemp(dir=subdir)
         if os.write(fd, contents) != len(contents):
@@ -238,6 +283,11 @@ class LogDB:
             self.remove_file(full_name)
         self._evict(full_name)
 
+    def remove_kpr(self, subdir, package, version):
+        full_name = os.path.join(subdir, self._kpr_name(package, version))
+        if os.path.exists(full_name):
+            self.remove_file(full_name)
+
     def stat(self, subdir, package, version):
         full_name = os.path.join(subdir, self._log_name(package, version))
         return os.stat(full_name)
@@ -251,6 +301,32 @@ class LogfileExists(Exception):
 
 class PackagesDB:
 
+    # these packages are uses as dependencies but are only available
+    # from foreign architectures
+    # HACK: this hardcoded list should be moved to some data file
+    _foreign_packages = {
+        "ia32-libs-i386": "i386",
+        "ia32-libs-gtk-i386": "i386",
+        "libnss-mdns-i386": "i386",
+    }
+
+    # these packages are used as dependencies but are only available on
+    # some architectures or from third-party repositories
+    # HACK: this hardcoded list should be moved to some data file
+    _ignored_missing_dependencies = [
+        # non-linux
+        "kbdcontrol",
+        "vidcontrol",
+        # jessie
+        "mediawiki",
+        # stretch
+        "llvm-3.7-dev",
+        # third-party
+        "oracle-instantclient-basic",
+        "oracle-instantclient11.2-basic",
+        "oracle-instantclient12.1-basic",
+    ]
+
     # keep in sync with piuparts-report.py: emphasize_reason()
     # FIXME: can we reorder this list or remove entries without breaking the counts.txt for the plot?
     _states = [
@@ -267,13 +343,17 @@ class PackagesDB:
         "unknown",
         "unknown-preferred-alternative",  # obsolete
         "no-dependency-from-alternatives-exists",  # obsolete
+        "outdated",
+        # "foreign:*",  # can only happen as query result for a dependency
         # "does-not-exist",  # can only happen as query result for a dependency
+        # "ignore-does-not-exist",  # can only happen as query result for a dependency
     ]
 
     _good_states = [
         "successfully-tested",
         "essential-required",
-    ]
+        "ignore-does-not-exist",
+    ] + ["foreign:%s" % arch for arch in set(_foreign_packages.values())]
 
     _obsolete_states = [
         "essential-required",
@@ -339,6 +419,7 @@ class PackagesDB:
         if recycle:
             self._recycle = pformat % recycle
             self._all.append(self._recycle)
+        self._most = [x for x in self._all if x not in [self._reserved, self._recycle]]
 
     def create_subdirs(self):
         for sdir in self._all:
@@ -372,10 +453,10 @@ class PackagesDB:
         pf2 = PackagesFile()
         pf2.load_packages_urls(urls)
         for package in self.get_all_packages():
-            if package["Package"] in pf2:
-                package["Version"] = pf2[package["Package"]]["Version"]
+            if package.name() in pf2:
+                package.set_test_versions(pf2[package.name()].version())
             else:
-                package["Version"] = "None"
+                package.set_test_versions("None")
 
     def get_urls(self):
         urls = []
@@ -429,15 +510,45 @@ class PackagesDB:
                     more += dep_pkg.dependencies()
         return circular
 
-    def _lookup_package_state(self, package):
-        if self._recycle_mode and self._logdb.log_exists(package, [self._recycle]):
-            return "unknown"
+    def _is_successfully_tested(self, package):
+        # a pass/ log exists but no corresponding recycle/ log exists
         if self._logdb.log_exists(package, [self._ok]):
-            return "successfully-tested"
+            if not (self._recycle_mode and self._logdb.log_exists(package, [self._recycle])):
+                return True
+        return False
+
+    def _lookup_package_state(self, package, use_cached_success, check_outdated):
+        if check_outdated:
+            # Check if dependency databases have a newer version of this package.
+            # Use the actual package versions, not the target versions.
+            curr_ver = package.version()
+            for db in self._dependency_databases:
+                dep_ver = db.get_version(package.name())
+                if dep_ver is not None and apt_pkg.version_compare(curr_ver, dep_ver) < 0:
+                    #logging.info("[%s] outdated: %s %s < %s @[%s]" % (self.prefix, package.name(), curr_ver, dep_ver, db.prefix))
+                    return "outdated";
+        if self._recycle_mode:
+            if self._logdb.log_exists(package, [self._reserved]):
+                return "waiting-to-be-tested"
+            if self._logdb.log_exists(package, [self._recycle]):
+                return "unknown"
+        if self._logdb.log_exists(package, [self._ok]):
+            success = True
+            if not use_cached_success:
+                # if a pass/ log exists but any dependency may be not
+                # trivially satisfiable do not skip dependency resolution
+                for dep in package.dependencies():
+                    if not self.get_package(dep, resolve_virtual=True):
+                        success = False
+                        break
+            if success:
+                return "successfully-tested"
         if self._logdb.log_exists(package, [self._fail] + self._morefail):
             return "failed-testing"
         if self._logdb.log_exists(package, [self._evil]):
             return "cannot-be-tested"
+        if self._logdb.log_exists(package, [self._reserved]):
+            return "waiting-to-be-tested"
         return "unknown"
 
     def _compute_package_state(self, package):
@@ -467,7 +578,7 @@ class PackagesDB:
                     prefer_alt_score = -1
                     prefer_alt = None
                     for alternative in alt_deps[d]:
-                        altdep_state = self.get_package_state(alternative)
+                        altdep_state = self.get_best_package_state(alternative)
                         if altdep_state != "does-not-exist":
                             alt_found += 1
                             if prefer_alt_score < 3 and altdep_state == "essential-required":
@@ -501,6 +612,8 @@ class PackagesDB:
                 testable = False
                 break
         if testable:
+            if self._is_successfully_tested(package):
+                return "successfully-tested"
             return "waiting-to-be-tested"
 
         # treat circular-dependencies as testable (for the part of the circle)
@@ -517,6 +630,8 @@ class PackagesDB:
                     testable = False
                     break
             if testable:
+                if self._is_successfully_tested(package):
+                    return "successfully-tested"
                 return "waiting-to-be-tested"
 
         for dep, dep_state in dep_states:
@@ -525,15 +640,7 @@ class PackagesDB:
 
         return "unknown"
 
-    def _compute_package_states(self):
-        if self._in_state is not None:
-            return
-
-        self._stamp = time.time()
-
-        for subdir in self._all:
-            self._logdb.bulk_load_dir(subdir)
-
+    def _initialize_package_states(self, use_cached_success, check_outdated):
         self._find_all_packages()
 
         self._package_state = {}
@@ -542,8 +649,8 @@ class PackagesDB:
             self._in_state[state] = []
         todo = []
 
-        for package_name, package in self._packages.iteritems():
-            state = self._lookup_package_state(package)
+        for package_name, package in six.iteritems(self._packages):
+            state = self._lookup_package_state(package, use_cached_success, check_outdated)
             assert state in self._states
             self._package_state[package_name] = state
             if state == "unknown":
@@ -551,8 +658,25 @@ class PackagesDB:
             else:
                 self._in_state[state].append(package_name)
 
+        return todo
+
+    def _compute_package_states(self, use_cached_success=False):
+        if self._in_state is not None:
+            return
+
+        self._stamp = time.time()
+
+        for subdir in self._all:
+            self._logdb.bulk_load_dir(subdir)
+
+        todo = self._initialize_package_states(use_cached_success=use_cached_success, check_outdated=False)
+
         for db in self._dependency_databases:
-            db._compute_package_states()
+            db._compute_package_states(use_cached_success=True)
+
+        if self._dependency_databases:
+            # redo the initialization to properly resolve "outdated" packages after the dependency databases have been initialized
+            todo = self._initialize_package_states(use_cached_success=use_cached_success, check_outdated=True)
 
         while todo:
             package_names = todo
@@ -567,7 +691,7 @@ class PackagesDB:
                     else:
                         self._in_state[state].append(package_name)
                         self._package_state[package_name] = state
-                        done.append(package)
+                        done.append(package_name)
             if not done:
                 # If we didn't do anything this time, we sure aren't going
                 # to do anything the next time either.
@@ -615,13 +739,13 @@ class PackagesDB:
     def get_version(self, name):
         self._find_all_packages()
         if name in self._packages:
-            return self._packages[name]["Version"]
+            return self._packages[name].version()
         return None
 
-    def get_source(self, name):
+    def get_test_versions(self, name):
         self._find_all_packages()
         if name in self._packages:
-            return self.get_control_header(name, "Source")
+            return self._packages[name].test_versions()
         return None
 
     def get_providers(self, name, recurse=True):
@@ -644,21 +768,7 @@ class PackagesDB:
 
     def get_control_header(self, package_name, header):
         self._find_all_packages()
-        if header == "Source":
-            # binary packages build from the source package with the same name
-            # don't have a Source header, so let's try:
-            try:
-                _source = self._packages[package_name][header]
-                # for binNMU the Source header in Packages files holds the version
-                # too, so we need to chop it of:
-                if " " in _source:
-                    source, version = _source.split(" ")
-                else:
-                    source = _source
-            except:
-                source = self._packages[package_name]["Package"]
-            return source
-        elif header == "Uploaders":
+        if header == "Uploaders":
             # not all (source) packages have an Uploaders header
             uploaders = ""
             try:
@@ -672,6 +782,11 @@ class PackagesDB:
     def get_package_state(self, package_name, resolve_virtual=True, recurse=True):
         self._compute_package_states()
         if package_name in self._package_state:
+            if recurse and self._package_state[package_name] == "outdated":
+                for db in self._dependency_databases:
+                    state = db.get_package_state(package_name, resolve_virtual=resolve_virtual, recurse=False)
+                    if state not in ["does-not-exist", "outdated"]:
+                        return state
             return self._package_state[package_name]
         if package_name in self._virtual_packages:
             if resolve_virtual:
@@ -684,9 +799,10 @@ class PackagesDB:
                 state = db.get_package_state(package_name, resolve_virtual=resolve_virtual, recurse=False)
                 if state != "does-not-exist":
                     return state
-        if package_name in ["ia32-libs-i386", "ia32-libs-gtk-i386", "libnss-mdns-i386"]:
-            # HACK! these are arch=i386 packages needed on amd64
-            return "essential-required"
+        if package_name in self._foreign_packages:
+            return "foreign:%s" % self._foreign_packages[package_name]
+        if package_name in self._ignored_missing_dependencies:
+            return "ignore-does-not-exist"
         return "does-not-exist"
 
     def get_best_package_state(self, package_name, resolve_virtual=True, recurse=True):
@@ -699,7 +815,7 @@ class PackagesDB:
         if not providers:
             return package_state
         states = [self.get_package_state(name, resolve_virtual=False, recurse=recurse) for name in [package_name] + providers]
-        for state in self._good_states + self._propagate_waiting_state.keys() + self._propagate_error_state.keys():
+        for state in self._good_states + list(self._propagate_waiting_state.keys()) + self._propagate_error_state.keys():
             if state in states:
                 return state
         return package_state
@@ -717,7 +833,7 @@ class PackagesDB:
             )
 
         try:
-            statobj = self._logdb.stat(self._recycle, p["Package"], p["Version"])
+            statobj = self._logdb.stat(self._recycle, p.name(), p.test_versions())
             ctime = statobj[stat.ST_CTIME]  # last inode modification = time of linking into recycle/
             mtime = statobj[stat.ST_MTIME]
         except OSError:
@@ -750,19 +866,21 @@ class PackagesDB:
         self._candidates_for_testing.remove(p)
 
     def reserve_package(self):
-        all_but_recycle = [x for x in self._all if x != self._recycle]
         for p in self._find_packages_ready_for_testing():
+            if self._logdb.log_exists(p, [self._reserved]):
+                self._remove_unavailable_candidate(p)
+                continue
             if self._recycle_mode and self._logdb.log_exists(p, [self._recycle]):
-                for vdir in all_but_recycle:
+                for vdir in [x for x in self._most if x != self._ok]:
                     if self._logdb.log_exists(p, [vdir]):
-                        self._logdb.remove(vdir, p["Package"], p["Version"])
-                        logging.info("Recycled %s %s %s" % (vdir, p["Package"], p["Version"]))
-            if self._logdb.log_exists(p, all_but_recycle):
+                        self._logdb.remove(vdir, p.name(), p.test_versions())
+                        logging.info("Recycled %s %s %s" % (vdir, p.name(), p.test_versions()))
+            elif self._logdb.log_exists(p, self._most):
                 self._remove_unavailable_candidate(p)
                 continue
             if self._logdb.log_exists(p, [self._recycle]):
-                self._logdb.remove(self._recycle, p["Package"], p["Version"])
-            if self._logdb.create(self._reserved, p["Package"], p["Version"], ""):
+                self._logdb.remove(self._recycle, p.name(), p.test_versions())
+            if self._logdb.create(self._reserved, p.name(), p.test_versions(), ""):
                 return p
         return None
 
@@ -774,37 +892,42 @@ class PackagesDB:
         with open(self._submissions, "a") as submissions:
             submissions.write("%d %s %s %s\n" % (time.time(), category, package, version))
 
+    def _remove_logs_if_reserved(self, package, version):
+        if self._logdb.log_exists2(package, version, [self._reserved]):
+            for vdir in self._most:
+                if self._logdb.log_exists2(package, version, [vdir]):
+                    self._logdb.remove(vdir, package, version)
+                    logging.info("Recycled %s %s %s" % (vdir, package, version))
+            self._logdb.remove(self._reserved, package, version)
+
     def unreserve_package(self, package, version):
         self._check_for_acceptability_as_filename(package)
         self._check_for_acceptability_as_filename(version)
+        if self._logdb.log_exists2(package, version, [self._reserved]):
+            if not self._logdb.log_exists2(package, version, [self._recycle]):
+                # restore possible recycle marker
+                if self._logdb.log_exists2(package, version, self._most):
+                    self._logdb.create(self._recycle, package, version, "")
         self._logdb.remove(self._reserved, package, version)
 
-    def pass_package(self, package, version, log):
+    def _process_log(self, package, version, log, subdir, result):
         self._check_for_acceptability_as_filename(package)
         self._check_for_acceptability_as_filename(version)
-        if self._logdb.create(self._ok, package, version, log):
-            self._logdb.remove(self._reserved, package, version)
-            self._record_submission("pass", package, version)
+        self._remove_logs_if_reserved(package, version)
+        if self._logdb.create(subdir, package, version, log):
+            self._record_submission(result, package, version)
+            self._logdb.remove_kpr(subdir, package, version)
         else:
-            raise LogfileExists(self._ok, package, version)
+            raise LogfileExists(subdir, package, version)
+
+    def pass_package(self, package, version, log):
+        self._process_log(package, version, log, self._ok, "pass")
 
     def fail_package(self, package, version, log):
-        self._check_for_acceptability_as_filename(package)
-        self._check_for_acceptability_as_filename(version)
-        if self._logdb.create(self._fail, package, version, log):
-            self._logdb.remove(self._reserved, package, version)
-            self._record_submission("fail", package, version)
-        else:
-            raise LogfileExists(self._fail, package, version)
+        self._process_log(package, version, log, self._fail, "fail")
 
     def make_package_untestable(self, package, version, log):
-        self._check_for_acceptability_as_filename(package)
-        self._check_for_acceptability_as_filename(version)
-        if self._logdb.create(self._evil, package, version, log):
-            self._logdb.remove(self._reserved, package, version)
-            self._record_submission("untestable", package, version)
-        else:
-            raise LogfileExists(self._evil, package, version)
+        self._process_log(package, version, log, self._evil, "untestable")
 
     def _get_rdep_dict(self):
         """Return dict of one-level reverse dependencies by package"""
